searchData={"items":[{"type":"behaviour","title":"NimblePool","doc":"`NimblePool` is a tiny resource-pool implementation.\n\nPools in the Erlang VM, and therefore in Elixir, are generally process-based: they manage a group of processes. The downside of said pools is that when they have to manage resources, such as sockets or ports, the additional process leads to overhead.\n\nIn such pools, you usually end-up with two scenarios:\n\n  * You invoke the pool manager, which returns the pooled process, which performs the operation on the socket or port for you, returning you the reply. This approach is non-optimal because all of the data sent and returned by the resource needs to be copied between processes\n\n  * You invoke the pool manager, which returns the pooled process, which gives you access to the resource. Then you can act directly on the resource, avoiding the data copying, but you need to keep the state of the resource in sync with the process\n\nNimblePool allows you to implement the second scenario without the addition of processes, which leads to a simpler and more efficient implementation. You should consider using NimblePool whenever you have to manage sockets, ports, or NIF resources and you want the client to perform one-off operations on them. For example, NimblePool is a good solution to manage HTTP/1 connections, ports that need to communicate with long-running programs, etc.\n\nThe downside of NimblePool is that, because all resources are under a single process, any resource management operation will happen on this single process, which is more likely to become a bottleneck. This can be addressed, however, by starting one NimblePool per scheduler and by doing scheduler-based dispatches.\n\nNimblePool may not be a good option to manage processes. After all, the goal of NimblePool is to avoid creating processes for resources. If you already have a process, using a process-based pool such as `poolboy` will provide a better abstraction.\n\nFinally, avoid using NimblePool to manage resources that support multiplexing, such as HTTP/2 connections. In fact, pools are not a good option to manage resources with multiplexing in general, as the pool removes the ability to multiplex.","ref":"NimblePool.html"},{"type":"behaviour","title":"Types of callbacks - NimblePool","doc":"NimblePool has two types of callbacks. Worker callbacks and pool callbacks. The worker callbacks configure the behaviour of each worker, such as initialization, checkin and checkout. The pool callbacks configure general pool behaviour, such as initialization and queueing.","ref":"NimblePool.html#module-types-of-callbacks"},{"type":"behaviour","title":"Examples - NimblePool","doc":"To use `NimblePool`, you must define a module that implements the pool worker logic, outlined in the `NimblePool` behaviour.\n\n#","ref":"NimblePool.html#module-examples"},{"type":"behaviour","title":"Port-based example - NimblePool","doc":"The example below keeps ports on the pool and check them out on every command. Please read the docs for `Port` before using the approach below, especially in regards to zombie ports.\n\n```elixir\ndefmodule PortPool do\n  @behaviour NimblePool\n\n  @doc ~S\"\"\"\n  Executes a given command against a port kept by the pool.\n\n  First we start a pool of ports:\n\n      iex> child = {NimblePool, worker: {PortPool, :cat}, name: PortPool}\n      iex> Supervisor.start_link([child], strategy: :one_for_one)\n\n  Now we can run commands against the ports in the pool:\n\n      iex> PortPool.command(PortPool, \"hello\\n\")\n      \"hello\\n\"\n      iex> PortPool.command(PortPool, \"world\\n\")\n      \"world\\n\"\n\n  \"\"\"\n  def command(pool, command, opts \\\\ []) do\n    pool_timeout = Keyword.get(opts, :pool_timeout, 5000)\n    receive_timeout = Keyword.get(opts, :receive_timeout, 15000)\n\n    NimblePool.checkout!(pool, :checkout, fn _from, port ->\n      send(port, {self(), {:command, command}})\n\n      receive do\n        {^port, {:data, data}} ->\n          try do\n            Process.unlink(port)\n            {data, :ok}\n          rescue\n            _ -> {data, :close}\n          end\n      after\n        receive_timeout ->\n          exit(:receive_timeout)\n      end\n    end, pool_timeout)\n  end\n\n  @impl NimblePool\n  def init_worker(:cat = pool_state) do\n    path = System.find_executable(\"cat\")\n    port = Port.open({:spawn_executable, path}, [:binary, args: [\"-\"]])\n    {:ok, port, pool_state}\n  end\n\n  @impl NimblePool\n  # Transfer the port to the caller\n  def handle_checkout(:checkout, {pid, _}, port, pool_state) do\n    Port.connect(port, pid)\n    {:ok, port, port, pool_state}\n  end\n\n  @impl NimblePool\n  # We got it back\n  def handle_checkin(:ok, _from, port, pool_state) do\n    {:ok, port, pool_state}\n  end\n\n  def handle_checkin(:close, _from, _port, pool_state) do\n    {:remove, :closed, pool_state}\n  end\n\n  @impl NimblePool\n  # On terminate, effectively close it\n  def terminate_worker(_reason, port, pool_state) do\n    Port.close(port)\n    {:ok, pool_state}\n  end\nend\n```\n\n#","ref":"NimblePool.html#module-port-based-example"},{"type":"behaviour","title":"HTTP/1-based example - NimblePool","doc":"The pool below uses [Mint](https://hexdocs.pm/mint) for HTTP/1 connections. It establishes connections eagerly. A better approach may be to establish connections lazily on checkout, as done by [Finch](https://github.com/keathley/finch), which is built on top of [Mint](https://github.com/elixir-mint/mint) + [NimbleOptions](https://github.com/dashbitco/nimble_options).\n\n```elixir\ndefmodule HTTP1Pool do\n  @behaviour NimblePool\n\n  @doc ~S\"\"\"\n  Executes a given command against a connection kept by the pool.\n\n  First we start the pool:\n\n      child = {NimblePool, worker: {HTTP1Pool, {:https, \"elixir-lang.org\", 443}}, name: HTTP1Pool}\n      Supervisor.start_link([child], strategy: :one_for_one)\n\n  Then we can use the connections in the pool:\n\n      iex> HTTP1Pool.get(HTTP1Pool, \"/\")\n      {:ok, %{status: 200, ...}}\n\n  \"\"\"\n  def get(pool, path, opts \\\\ []) do\n    pool_timeout = Keyword.get(opts, :pool_timeout, 5000)\n    receive_timeout = Keyword.get(opts, :receive_timeout, 15000)\n\n    NimblePool.checkout!(\n      pool,\n      :checkout,\n      fn _from, conn ->\n        {{kind, result_or_error}, conn} =\n          with {:ok, conn, ref} <- Mint.HTTP1.request(conn, \"GET\", path, [], nil),\n               {:ok, conn, result} <- receive_response([], conn, ref, %{}, receive_timeout) do\n            {{:ok, result}, transfer_if_open(conn)}\n          end\n\n        {{kind, result_or_error}, conn}\n      end,\n      pool_timeout\n    )\n  end\n\n  defp transfer_if_open(conn) do\n    if Mint.HTTP1.open?(conn) do\n      {:ok, conn}\n    else\n      :closed\n    end\n  end\n\n  defp receive_response([], conn, ref, response, timeout) do\n    {:ok, conn, entries} = Mint.HTTP1.recv(conn, 0, timeout)\n    receive_response(entries, conn, ref, response, timeout)\n  end\n\n  defp receive_response([entry | entries], conn, ref, response, timeout) do\n    case entry do\n      {kind, ^ref, value} when kind in [:status, :headers] ->\n        response = Map.put(response, kind, value)\n        receive_response(entries, conn, ref, response, timeout)\n\n      {:data, ^ref, data} ->\n        response = Map.update(response, :data, data, &(&1 <> data))\n        receive_response(entries, conn, ref, response, timeout)\n\n      {:done, ^ref} ->\n        {:ok, conn, response}\n\n      {:error, ^ref, error} ->\n        {:error, conn, error}\n    end\n  end\n\n  @impl NimblePool\n  def init_worker({scheme, host, port} = pool_state) do\n    parent = self()\n\n    async = fn ->\n      # TODO: Add back-off\n      {:ok, conn} = Mint.HTTP1.connect(scheme, host, port, [])\n      {:ok, conn} = Mint.HTTP1.controlling_process(conn, parent)\n      conn\n    end\n\n    {:async, async, pool_state}\n  end\n\n  @impl NimblePool\n  # Transfer the conn to the caller.\n  # If we lost the connection, then we remove it to try again.\n  def handle_checkout(:checkout, _from, conn, pool_state) do\n    with {:ok, conn} <- Mint.HTTP1.set_mode(conn, :passive) do\n      {:ok, conn, conn, pool_state}\n    else\n      _ -> {:remove, :closed, pool_state}\n    end\n  end\n\n  @impl NimblePool\n  # We got it back.\n  def handle_checkin(state, _from, _old_conn, pool_state) do\n    with {:ok, conn} <- state,\n         {:ok, conn} <- Mint.HTTP1.set_mode(conn, :active) do\n      {:ok, conn, pool_state}\n    else\n      {:error, _} -> {:remove, :closed, pool_state}\n    end\n  end\n\n  @impl NimblePool\n  # If it is closed, drop it.\n  def handle_info(message, conn) do\n    case Mint.HTTP1.stream(conn, message) do\n      {:ok, _, _} -> {:ok, conn}\n      {:error, _, _, _} -> {:remove, :closed}\n      :unknown -> {:ok, conn}\n    end\n  end\n\n  @impl NimblePool\n  # On terminate, effectively close it.\n  # This will succeed even if it was already closed or if we don't own it.\n  def terminate_worker(_reason, conn, pool_state) do\n    Mint.HTTP1.close(conn)\n    {:ok, pool_state}\n  end\nend\n```","ref":"NimblePool.html#module-http-1-based-example"},{"type":"function","title":"NimblePool.checkout!/4","doc":"Checks out a worker from the pool.\n\nIt expects a command, which will be passed to the `c:handle_checkout/4`\ncallback. The `c:handle_checkout/4` callback will return a client state,\nwhich is given to the `function`.\n\nThe `function` receives two arguments, the request\n(`{pid(), reference()}`) and the `client_state`.\nThe function must return a two-element tuple, where the first element is the\nreturn value for `checkout!/4`, and the second element is the updated `client_state`,\nwhich will be given as the first argument to `c:handle_checkin/4`.\n\n`checkout!/4` also has an optional `timeout` value. This value will be applied\nto the checkout operation itself. The \"check in\" operation happens asynchronously.","ref":"NimblePool.html#checkout!/4"},{"type":"function","title":"NimblePool.child_spec/1","doc":"Defines a pool to be started under the supervision tree.\n\nIt accepts the same options as `start_link/1` with the\naddition or `:restart` and `:shutdown` that control the\n\"Child Specification\".","ref":"NimblePool.html#child_spec/1"},{"type":"function","title":"Examples - NimblePool.child_spec/1","doc":"NimblePool.child_spec(worker: {__MODULE__, :some_arg}, restart: :temporary)","ref":"NimblePool.html#child_spec/1-examples"},{"type":"callback","title":"NimblePool.handle_cancelled/2","doc":"Handle cancelled checkout requests.\n\nThis callback is executed when a checkout request is cancelled unexpectedly.\n\nThe context argument may be `:queued` or `:checked_out`:\n\n* `:queued` means the cancellation happened before resource checkout. This may happen\nwhen the pool is starving under load and can not serve resources.\n\n* `:checked_out` means the cancellation happened after resource checkout. This may happen\nwhen the function given to `checkout!/4` raises.\n\nThis callback is optional.","ref":"NimblePool.html#c:handle_cancelled/2"},{"type":"callback","title":"NimblePool.handle_checkin/4","doc":"Checks a worker back in the pool.\n\nIt receives the potentially-updated `client_state`, returned by the `checkout!/4`\nanonymous function, and it must return either\n`{:ok, worker_state, pool_state}` or `{:remove, reason, pool_state}`.\n\n> ##","ref":"NimblePool.html#c:handle_checkin/4"},{"type":"callback","title":"Blocking the pool {: .warning} - NimblePool.handle_checkin/4","doc":">\n> This callback is synchronous and therefore will block the pool.\n> Avoid performing long work in here, instead do as much work as\n> possible on the client.\n\nOnce the connection is checked in, it may immediately be handed\nto another client, without traversing any of the messages in the\npool inbox.\n\nThis callback is optional.","ref":"NimblePool.html#c:handle_checkin/4-blocking-the-pool-warning"},{"type":"callback","title":"NimblePool.handle_checkout/4","doc":"Checks a worker out.\n\nThe `maybe_wrapped_command` is the `command` passed to `checkout!/4` if the worker\ndoesn't implement the `c:handle_enqueue/2` callback, otherwise it's the possibly-wrapped\ncommand returned by `c:handle_enqueue/2`.\n\nThis callback must return one of:\n\n  * `{:ok, client_state, worker_state, pool_state}` — the client state is given to\n    the callback function passed to `checkout!/4`. `worker_state` and `pool_state`\n    can potentially update the state of the checked-out worker and the pool.\n\n  * `{:remove, reason, pool_state}` — `NimblePool` will remove the checked-out worker and\n    attempt to checkout another worker.\n\n  * `{:skip, Exception.t(), pool_state}` — `NimblePool` will skip the checkout, the client will\n    raise the returned exception, and the worker will be left ready for the next\n    checkout attempt.\n\n> ##","ref":"NimblePool.html#c:handle_checkout/4"},{"type":"callback","title":"Blocking the pool {: .warning} - NimblePool.handle_checkout/4","doc":">\n> This callback is synchronous and therefore will block the pool.\n> Avoid performing long work in here. Instead, do as much work as\n> possible on the client.\n\nOnce the worker is checked out, the worker won't handle any\nmessages targeted to `c:handle_info/2`.","ref":"NimblePool.html#c:handle_checkout/4-blocking-the-pool-warning"},{"type":"callback","title":"NimblePool.handle_enqueue/2","doc":"Executed by the pool whenever a request to check out a worker is enqueued.\n\nThe `command` argument should be treated as an opaque value, but it can be\nwrapped with some data to be used in `c:handle_checkout/4`.\n\nIt must return either `{:ok, maybe_wrapped_command, pool_state}` or\n`{:skip, Exception.t(), pool_state}` if checkout is to be skipped.\n\n> ##","ref":"NimblePool.html#c:handle_enqueue/2"},{"type":"callback","title":"Blocking the pool {: .warning} - NimblePool.handle_enqueue/2","doc":">\n> This callback is synchronous and therefore will block the pool.\n> Avoid performing long work in here.\n\nThis callback is optional.","ref":"NimblePool.html#c:handle_enqueue/2-blocking-the-pool-warning"},{"type":"callback","title":"Examples - NimblePool.handle_enqueue/2","doc":"@impl NimblePool\n    def handle_enqueue(command, pool_state) do\n      {:ok, {:wrapped, command}, pool_state}\n    end","ref":"NimblePool.html#c:handle_enqueue/2-examples"},{"type":"callback","title":"NimblePool.handle_info/2","doc":"Receives a message in the pool and handles it as each worker.\n\nIt receives the `message` and it must return either\n`{:ok, worker_state}` to update the worker state, or `{:remove, reason}` to\nremove the worker.\n\nSince there is only a single pool process that can receive messages, this\ncallback is executed once for every worker when the pool receives `message`.\n\n> ##","ref":"NimblePool.html#c:handle_info/2"},{"type":"callback","title":"Blocking the pool {: .warning} - NimblePool.handle_info/2","doc":">\n> This callback is synchronous and therefore will block the pool while it\n> executes for each worker. Avoid performing long work in here.\n\nThis callback is optional.","ref":"NimblePool.html#c:handle_info/2-blocking-the-pool-warning"},{"type":"callback","title":"NimblePool.handle_ping/2","doc":"Handle pings due to inactivity on the worker.\n\nExecuted whenever the idle worker periodic timer verifies that a worker has been idle\non the pool for longer than the `:worker_idle_timeout` pool configuration (in milliseconds).\n\nThis callback must return one of the following values:\n\n  * `{:ok, worker_state}`: Updates worker state.\n\n  * `{:remove, user_reason}`: The pool will proceed to the standard worker termination\n      defined in `terminate_worker/3`.\n\n  * `{:stop, user_reason}`: The entire pool process will be terminated, and `terminate_worker/3`\n      will be called for every worker on the pool.\n\nThis callback is optional.","ref":"NimblePool.html#c:handle_ping/2"},{"type":"callback","title":"Max idle pings - NimblePool.handle_ping/2","doc":"The `:max_idle_pings` pool option is useful to prevent sequential termination of a large number\nof workers. However, it is important to keep in mind the following behaviours whenever\nutilizing it.\n\n  * If you are not terminating workers with `c:handle_ping/2`, you may end up pinging only\n    the same workers over and over again because each cycle will ping only the first\n    `:max_idle_pings` workers.\n\n  * If you are terminating workers with `c:handle_ping/2`, the last worker may be terminated\n    after up to `worker_idle_timeout + worker_idle_timeout * ceil(number_of_workers/max_idle_pings)`,\n    instead of `2 * worker_idle_timeout` milliseconds of idle time.\n\nFor instance consider a pool with 10 workers and a ping of 1 second.\n\nGiven a negligible worker termination time and a worst-case scenario where all the workers\ngo idle right after a verification cycle is started, then without `max_idle_pings` the\nlast worker will be terminated in the next cycle (2 seconds), whereas with a\n`max_idle_pings` of 2 the last worker will be terminated only in the 5th cycle (6 seconds).","ref":"NimblePool.html#c:handle_ping/2-max-idle-pings"},{"type":"callback","title":"Disclaimers - NimblePool.handle_ping/2","doc":"* On lazy pools, if no worker is currently on the pool the callback will never be called.\n    Therefore you can not rely on this callback to terminate empty lazy pools.\n\n  * On not lazy pools, if you return `{:remove, user_reason}` you may end up\n    terminating and initializing workers at the same time every idle verification cycle.\n\n  * On large pools, if many resources go idle at the same cycle, you may end up terminating\n    a large number of workers sequentially, which could lead to the pool being unable to\n    fulfill requests. See `:max_idle_pings` option to prevent this.","ref":"NimblePool.html#c:handle_ping/2-disclaimers"},{"type":"callback","title":"NimblePool.handle_update/3","doc":"Handles an update instruction from a checked out worker.\n\nSee `update/2` for more information.\n\nThis callback is optional.","ref":"NimblePool.html#c:handle_update/3"},{"type":"callback","title":"NimblePool.init_pool/1","doc":"Initializes the pool.\n\nIt receives the worker argument passed to `start_link/1` and must\nreturn `{:ok, pool_state}` upon successful initialization,\n`:ignore` to exit normally, or `{:stop, reason}` to exit with `reason`\nand return `{:error, reason}`.\n\nThis is a good place to perform a registration, for example.\n\nIt must return the `pool_state`. The `pool_state` is given to\n`init_worker`. By default, it simply returns the given arguments.\n\nThis callback is optional.","ref":"NimblePool.html#c:init_pool/1"},{"type":"callback","title":"Examples - NimblePool.init_pool/1","doc":"@impl NimblePool\n    def init_pool(options) do\n      Registry.register(options[:registry], :some_key, :some_value)\n    end","ref":"NimblePool.html#c:init_pool/1-examples"},{"type":"callback","title":"NimblePool.init_worker/1","doc":"Initializes the worker.\n\nIt receives the worker argument passed to `start_link/1` if `c:init_pool/1` is\nnot implemented, otherwise the pool state returned by `c:init_pool/1`. It must\nreturn `{:ok, worker_state, pool_state}` or `{:async, fun, pool_state}`, where the `fun`\nis a zero-arity function that must return the worker state.\n\nIf this callback returns `{:async, fun, pool_state}`, `fun` is executed in a **separate\none-off process**. Because of this, if you start resources that the pool needs to \"own\",\nyou need to transfer ownership to the pool process. For example, if your async `fun`\nopens a `:gen_tcp` socket, you'll have to use `:gen_tcp.controlling_process/2` to transfer\nownership back to the pool.\n\n> ##","ref":"NimblePool.html#c:init_worker/1"},{"type":"callback","title":"Blocking the pool {: .warning} - NimblePool.init_worker/1","doc":">\n> This callback is synchronous and therefore will block the pool, potentially\n> for a significant amount of time since it's executed in the pool process once\n> per worker. > If you need to perform long initialization, consider using the\n> `{:async, fun, pool_state}` return type.","ref":"NimblePool.html#c:init_worker/1-blocking-the-pool-warning"},{"type":"function","title":"NimblePool.start_link/1","doc":"Starts a pool.","ref":"NimblePool.html#start_link/1"},{"type":"function","title":"Options - NimblePool.start_link/1","doc":"* `:worker` - a `{worker_mod, worker_init_arg}` tuple with the worker\n    module that implements the `NimblePool` behaviour and the worker\n    initial argument. This argument is **required**.\n\n  * `:pool_size` - how many workers in the pool. Defaults to `10`.\n\n  * `:lazy` - When `true`, workers are started lazily, only when necessary.\n    Defaults to `false`.\n\n  * `:worker_idle_timeout` - Timeout in milliseconds to tag a worker as idle.\n    If not nil, starts a periodic timer on the same frequency that will ping\n    all idle workers using `handle_ping/2` optional callback .\n    Defaults to no timeout.\n\n  * `:max_idle_pings` - Defines a limit to the number of workers that can be pinged\n    for each cycle of the `handle_ping/2` optional callback.\n    Defaults to no limit. See `handle_ping/2` for more details.","ref":"NimblePool.html#start_link/1-options"},{"type":"function","title":"NimblePool.stop/3","doc":"Stops the given `pool`.\n\nThe pool exits with the given `reason`. The pool has `timeout` milliseconds\nto terminate, otherwise it will be brutally terminated.","ref":"NimblePool.html#stop/3"},{"type":"function","title":"Examples - NimblePool.stop/3","doc":"NimblePool.stop(pool)\n    #=> :ok","ref":"NimblePool.html#stop/3-examples"},{"type":"callback","title":"NimblePool.terminate_pool/2","doc":"Handle pool termination.\n\nThe `reason` argmument is the same given to GenServer's terminate/2 callback.\n\nIt is not necessary to terminate workers here because the\n`terminate_worker/3` callback has already been invoked.\n\nThis should be used only for clean up extra resources that can not be\nhandled by `terminate_worker/3` callback.\n\nThis callback is optional.","ref":"NimblePool.html#c:terminate_pool/2"},{"type":"callback","title":"NimblePool.terminate_worker/3","doc":"Terminates a worker.\n\nThe `reason` argument is:\n\n  * `:DOWN` whenever the client link breaks\n  * `:timeout` whenever the client times out\n  * one of `:throw`, `:error`, `:exit` whenever the client crashes with one\n    of the reasons above.\n  * `reason` if at any point you return `{:remove, reason}`\n  * if any callback raises, the raised exception will be given as `reason`.\n\nIt receives the latest known `worker_state`, which may not\nbe the latest state. For example, if a client checks out the\nstate and crashes, we don't fully know the `client_state`,\nso the `c:terminate_worker/3` callback needs to take such scenarios\ninto account.\n\nThis callback must always return `{:ok, pool_state}` with the potentially-updated\npool state.\n\nThis callback is optional.","ref":"NimblePool.html#c:terminate_worker/3"},{"type":"function","title":"NimblePool.update/2","doc":"Sends an **update** instruction to the pool about the checked out worker.\n\nThis must be called inside the `checkout!/4` callback function with\nthe `from` value given to `c:handle_checkout/4`.\n\nThis is useful to update the pool's state before effectively\nchecking the state in, which is handy when transferring\nresources requires two steps.","ref":"NimblePool.html#update/2"},{"type":"type","title":"NimblePool.client_state/0","doc":"","ref":"NimblePool.html#t:client_state/0"},{"type":"type","title":"NimblePool.from/0","doc":"","ref":"NimblePool.html#t:from/0"},{"type":"type","title":"NimblePool.init_arg/0","doc":"","ref":"NimblePool.html#t:init_arg/0"},{"type":"type","title":"NimblePool.pool/0","doc":"","ref":"NimblePool.html#t:pool/0"},{"type":"type","title":"NimblePool.pool_state/0","doc":"","ref":"NimblePool.html#t:pool_state/0"},{"type":"type","title":"NimblePool.user_reason/0","doc":"","ref":"NimblePool.html#t:user_reason/0"},{"type":"type","title":"NimblePool.worker_state/0","doc":"","ref":"NimblePool.html#t:worker_state/0"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,49,46,50]}}